#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/fanotify.h>
#include <sys/types.h>
#include <unistd.h>

int read_all(int fd, void *buffer, size_t length)
{
    ssize_t readin;

    while (length > 0)
    {
        readin = read(fd, buffer, length);
        if (readin <= 0)
        {
            return 0;
        }
        length -= readin;
        buffer += readin;
    }
    return 1;
}

int file_is_infected(int fd)
{
    const char *eicar =
        "X5O!P%@AP[4\\PZX54(P^)7CC)7}$EICAR-STANDARD-ANTIVIRUS-TEST-FILE!$H+H*";
    const size_t length = strlen(eicar);
    char buffer[length];
    off_t filesize;

    if (!read_all(fd, buffer, length))
    {
        return 0;
    }
    if (memcmp(buffer, eicar, length) != 0)
    {
        return 0;
    }
    filesize = lseek(fd, 0, SEEK_END);
    if (filesize < 0 || filesize > 128)
    {
        return 0;
    }
    return 1;
}

void print_filename(int fd)
{
    char path[1024];
    int path_len;

    sprintf(path, "/proc/self/fd/%d", fd);
    path_len = readlink(path, path, sizeof(path)-1);
    if (path_len < 0)
    {
        puts("?");
    }
    else
    {
        path[path_len] = 0;
        puts(path);
    };
}

int main(int argc, char *argv[])
{
    int fanotify_fd;
    struct fanotify_event_metadata event;
    struct fanotify_response response;

    fanotify_fd = fanotify_init(FAN_CLASS_CONTENT, 0);
    if (fanotify_fd == -1)
    {
        perror("fanotify_init");
        return 1;
    }
    if (fanotify_mark(fanotify_fd, FAN_MARK_ADD | FAN_MARK_MOUNT,
        FAN_OPEN_PERM, AT_FDCWD, "/") == -1)
    {
        perror("fanotify_mark");
        return 1;
    }
    while (read_all(fanotify_fd, &event, sizeof(event)))
    {
        if (event.event_len != sizeof(event) ||
            event.vers != FANOTIFY_METADATA_VERSION)
        {
            puts("Unexpected event format.");
            return 1;
        }
        if ((event.mask & FAN_OPEN_PERM) == 0)
        {
            puts("Unexpected event type.");
            return 1;
        }
        response.fd = event.fd;
        if (file_is_infected(event.fd))
        {
            response.response = FAN_DENY;
            printf("INFECTED ");
        }
        else
        {
            response.response = FAN_ALLOW;
            printf("CLEAN    ");
        }
        print_filename(event.fd);
        if (write(fanotify_fd, &response, sizeof(response)) == -1)
        {
            perror("write");
            return 1;
        }
        close(event.fd);
    }
    puts("Read error.");
    return 1;
}
